/******************************************************************************
 *
 *     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
 *     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
 *     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
 *     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
 *     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
 *     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
 *     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
 *     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
 *     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
 *     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
 *     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
 *     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *     FOR A PARTICULAR PURPOSE.
 *     
 *     (c) Copyright 2005 Xilinx, Inc.
 *     All rights reserved.
 *
 ******************************************************************************/

//----------------------------------------------------------------------------
// Filename:		main.c
// 
// Description:		The top level file in the Mandelbrot image generation project 
//					It first runs and times the software-only project 
//					Subsequently it runs and times the accelerated hardware 
//					version
//
// Design Notes:
//   
//-----------------------------------------------------------------------------
// Xilinx APD
// History:
// July 2005 - Created - Kunal Shenoy
//
//-----------------------------------------------------------------------------

// Headers requried for desktop simulation 
#ifndef IMPULSE_C_TARGET
	#include <stdio.h>
	#include <stdlib.h>
#endif

//co.h contains the headers for CoDeveloper functions
#include "co.h"
//mand.h contains the image parameters and useful macros 
#include "mand.h"
#define TFT_DEVICE_ID 0
//TBD : get errors when included
//if not included get warning for implicit declaration of printfs
//#include <stdio.h>

#ifdef IMPULSE_C_TARGET //If compiled for FPGA
	//FPGA specific header files
	#include "bootload_basicgraphics.h"
	#include "initializeDisplay.h"
	#include "xtft_main.h"
	#include "xparameters.h"
	#include "xcache_l.h"
	#include "xgpio_l.h"
	#include "sleep.h"
	#include "string.h"
	#include "stop_watch.h"
	#include "mand_sw_only.h"
	#include "gpio_lcd_led.h"
#else //If compiled for desktop simulation
	FILE * outfileHW;	//file handle for hardware BMP file
	FILE * outfileSW; 	//file handle for software BMP file
	//Bitmap filenames to store filenames of desktop simulations
	//Run simulations and check these files for images
	#define OUTPUT_FILE_HW "mandelbrot_hw.bmp"
	#define OUTPUT_FILE_SW "mandelbrot_sw.bmp"
	
	#include "bmp.h"	
	#include "mand_sw_only.h"
#endif

//co_initialize() is defined in mand_accel_hw.c
extern co_architecture co_initialize(void *);

//Top level function of project
int main()
{
	co_architecture my_arch; 			//CoDeveloper system
	double xmax,xmin,ymax,ymin,dx,dy; 	//image parameters
	//char line1[17], line2[17];			// LCD Output buffers
	static double theFactor = 0; 		//factor to change image parameters 
										//for demo

	//If compiled for FPGA declare time-keeping variables
	//and enable caches
#ifdef IMPULSE_C_TARGET
	char time_total_string[20];
	uint32 time_total;
	uint32 speedup;
	uint first_pass = 1;

	//enable caches for DDR and BRAMs
  	XCache_EnableICache(0x80000001);
  	XCache_EnableDCache(0x80000001);

#ifdef XPAR_LCD_7BIT_GPIO_BASEADDR
	// initialize the LCD
	LCD_Init();
	// write welcome message to display
	strcpy (line1, "Mandelbrot Demo");
	strcpy (line2, "APU + C-2-HDL");
	LCD_Write (line1, line2);
#endif

	//The infinite while loop is used to display multiple images when running
	//the demo on FPGA.
	while(1)
	{
#endif
		printf("Entered Main()\n\r");

#ifdef IMPULSE_C_TARGET //If compiled for FPGA
	
#ifdef XPAR_LEDS_4BIT_BASEADDR
		//Turn on LEDs on board to show that demo is running
		WriteToGPOutput(XPAR_LEDS_4BIT_BASEADDR, 4);
#endif

#ifdef XPAR_LEDS_8BIT_BASEADDR
		//Turn on LEDs on board to show that demo is running
		WriteToGPOutput(XPAR_LEDS_8BIT_BASEADDR, 8);
#endif

		// Initialize TFT display 
		XTft_Initialize(&Tft, TFT_DEVICE_ID);
		//Sleep inserted to compensate for delay 
		//in montior warming-up after startup
		usleep(2000000); 
		initializeDisplay();

#else	//If compiled for desktop simulation
		//bitmap file initialization 
		BITMAPFILEHEADER header;  // Bitmap header
		BITMAPINFO info;          // Bitmap information
		const char * FileNameHW = OUTPUT_FILE_HW;
		const char * FileNameSW = OUTPUT_FILE_SW; 

		//Open HW and SW bitmap files
		outfileHW = fopen(FileNameHW, "wb");
		if ( outfileHW == NULL ) 
		{
			fprintf(stderr, "Error opening BMP file %s for writing\n", FileNameHW);
			exit(-1);
		}

		outfileSW = fopen(FileNameSW, "wb");
		if ( outfileSW == NULL ) 
		{
			fprintf(stderr, "Error opening BMP file %s for writing\n", FileNameSW);
			exit(-1);
		}
		// Write the file header and info sections of bitmap file
		//XSIZE and YSIZE defined in mand.h
		header.bfType = 0x4d42; // BMP file
		header.bfSize = XSIZE * YSIZE + sizeof(header) + sizeof(info);
		header.bfOffBits = sizeof(header) + sizeof(info);
		info.bmiHeader.biSize = 40;
		info.bmiHeader.biWidth = XSIZE;
		info.bmiHeader.biHeight = YSIZE;
		info.bmiHeader.biPlanes = 1;
		info.bmiHeader.biBitCount = 24;
		info.bmiHeader.biCompression = 0;
		info.bmiHeader.biSizeImage = XSIZE * YSIZE;
		info.bmiHeader.biXPelsPerMeter = 11811;
		info.bmiHeader.biYPelsPerMeter = 11811;
		info.bmiHeader.biClrUsed = 0;
		info.bmiHeader.biClrImportant = 0;

		//Write file headers to HW-result file
		if (fwrite(&header, 1, sizeof(BITMAPFILEHEADER), outfileHW) < 1)
		{
			fclose(outfileHW);
			fprintf(stderr, "Error writing BMP header for %s\n", FileNameHW);
			exit(-1);
		}
		if (fwrite(&info, 1, sizeof(BITMAPINFO), outfileHW) < 1)
		{
			fclose(outfileHW);
			fprintf(stderr, "Error writing BMP info for %s\n", FileNameHW);
			exit(-1);
		}
	
		//Write file headers to SW-result file
		if (fwrite(&header, 1, sizeof(BITMAPFILEHEADER), outfileSW) < 1)
		{
			fclose(outfileSW);
			fprintf(stderr, "Error writing BMP header for %s\n", FileNameSW);
			exit(-1);
		}
		if (fwrite(&info, 1, sizeof(BITMAPINFO), outfileSW) < 1)
		{
			fclose(outfileSW);
			fprintf(stderr, "Error writing BMP info for %s\n", FileNameSW);
			exit(-1);
		}
#endif

////////////////////////////////////////////////////
//             SW-only Version                    //
////////////////////////////////////////////////////

		printf("\nSW-only Version\n\r");
	
		//calculate image parameters
		xmin = XL - theFactor;
		xmax = XH - theFactor;
		ymin = YL - theFactor;
		dx = (XH -XL) / XSIZE;
		dy = dx; //(YH-YL)/YSIZE;
		ymax=YL + dy*YSIZE;

		if (theFactor < 0.036)
			theFactor += 0.003;
		else
			theFactor = 0;

#ifdef IMPULSE_C_TARGET
		flashPowerpc();
		//start timer to keep track of completion time
		stop_watch(1, 0);
#endif

		//Run SW only version
		fix_mandelbrot((int32) (xmax*(1<<FRACBITS)),
						(int32) (xmin*(1<<FRACBITS)),
						(int32) (ymax*(1<<FRACBITS)),
						(int32) (ymin*(1<<FRACBITS)),
					   (uint32) (dx*(1<<FRACBITS)),
					   (uint32) (dy*(1<<FRACBITS)),
					   (uint32) (dx*(1<<(FRACBITS-1))),
					   (uint32) (dy*(1<<(FRACBITS-1))));

		
#ifdef IMPULSE_C_TARGET
		//Calculate and display completion time for SW
		speedup = stop_watch(5, 0) / 1000;
		sprintf(time_total_string, " %d ms", speedup);
		XTft_SetPos(&Tft, 146, 175);
		XTft_WriteColorString(&Tft, time_total_string, 0xFF0000, 0x00000000);
#endif 
		printf("\nSW done\n\n\r");

///////////////////////////////////////////////////
//HW-SW Version
///////////////////////////////////////////////////
		printf("APU + C-to-HDL (HW) Version\n\r");

#ifdef IMPULSE_C_TARGET //initialize the CoDeveloper system only once
		if (first_pass)
#endif
			//Usage: (co_architecture)arch_name = co_initialize(paramters_to_be_sent_to_processes);
			//Every CoDeveloper project must havea initialization function. It provides a 
			//define starting point for Impulse C compilation
			my_arch = co_initialize(NULL);

#ifdef IMPULSE_C_TARGET //If running on FPGA
		flashApu();
		stop_watch(1, 0);	//keep track of completion time
#endif	
		//Run HW version
		co_execute(my_arch);

#ifdef IMPULSE_C_TARGET //If running on FPGA
		//Calculate and display completion time for HW
		time_total = stop_watch(5, 0) / 1000;
		if(time_total)
		{
			sprintf(time_total_string, " %d ms", time_total);
			XTft_SetPos(&Tft, 146, 190);
			XTft_WriteColorString(&Tft, time_total_string, 0x0FF00, 0x00000000);
		}
#endif
		printf("\nHW done\n\r");

		//calculate and plot speedup comparison between software-only and hardware-software executions
#ifdef IMPULSE_C_TARGET
		speedup /= time_total;
		if(time_total)
		{
			sprintf(time_total_string, " %dX", speedup);
			XTft_SetPos(&Tft, 146, 220);
			XTft_WriteColorString(&Tft, time_total_string, 0xFF00, 0x00000000);
		}
		//wait 10 seconds after comparison is completed
		//to allow users to see speedup
		usleep(10000000);	
		first_pass = 0;
	}
#else
	fclose(outfileHW);
	fclose(outfileSW);
#endif
	printf("\nExited Main()\n\r");
	IF_SIM(printf("Press any key to continue...\n\r");)
	fflush(stdout);
	IF_SIM(getc(stdin));
  	return 0;
}

