// Copyright(c) 2003-2005 Impulse Accelerated Technologies, Inc.  All rights reserved.
// www.ImpulseC.com
//  
// This source file may be used and distributed without restriction provided that this
// copyright notice is not removed from the file and that any derivative work contains
// this copyright notice.
//

#include <stdio.h>
#include "co.h"
#include "mand.h"

//mand_accel_sw is in mand_accel_sw.c and is referenced here in mand_config()
//where all the processes in the system are declared and defined
extern void mand_accel_sw(co_stream config_stream, co_stream pixel_stream);

//mand_accel_hw() is the mandelbrot generation function to be converted to a HW process.
//config_stream = stream over which mand_accel_hw() reads in parameters from mand_accel_sw().
//pixel_stream = stream over which mand_accel_hw() sends pixel values to mand_accel_sw().
//mand_accel_hw takes in image parameters from mand_accel_sw, calculates pixel values
//and sends it back to mand_accel_sw.
void mand_accel_hw(co_stream config_stream, co_stream pixel_stream) 
{
    co_int32 xmax, xmin, ymax, ymin, dx, dy; //image parameters
    co_uint24 B, G, R, BGR; //pixel color variables
    co_int32 i, j, k; //loop variables
    co_int32 c_imag, c_real; //complex constant number added to product of squaring
	co_int32 two, four; //used to hold integer constants in fixed-point format
    co_int32 result, tmp; //calculation variables
    co_int32 z_real, z_imag; //complex number used in calculation

	//constant integers assigned in fixed point format
    two = FXCONST(2);
    four = FXCONST(4);

	//co_stream_open() resets the internal state of the stream
	//Usage: co_stream_open(stream_name, O_RDONLY/0_WRONLY, size_of_stream_element);
	//Streams are uni-directional and hence they have to be opened to be read
	//or written to exclusively
    co_stream_open(config_stream, O_RDONLY, INT_TYPE(32));

    //Read in parameters of Mandelbrot image from sw function over the config_stream stream.
	//Function will return error code and loop will exit, if SW process on other end writes
	//an end-of-stream token.
	//co_stream_read() reads data from a non-empty stream or else it blocks.
	//Usage: co_stream_read(stream_name, variable_address, data_read_size);
    while (co_stream_read(config_stream, &xmax, sizeof(co_int32)) == co_err_none) 
	{
        co_stream_read(config_stream, &xmin, sizeof(co_int32));
        co_stream_read(config_stream, &ymax, sizeof(co_int32));
        co_stream_read(config_stream, &ymin, sizeof(co_int32));
        co_stream_read(config_stream, &dx, sizeof(co_int32));
        co_stream_read(config_stream, &dy, sizeof(co_int32));

		//co_stream_open() resets the internal state of the stream
		//Usage: co_stream_open(stream_name, O_RDONLY/0_WRONLY, size_of_stream_element);
        co_stream_open(pixel_stream, O_WRONLY, UINT_TYPE(24));
        c_imag = ymax;
        for (j = 0; j < YSIZE; j++) //For each horizontal line
		{
            c_real=xmin;
			for (i=0; i<XSIZE; i++) //For each pixel of each horizontal line
			{
                z_real = z_imag = 0;
                k = 0;
				//Calculate pixel color
				do 
				{
//Hardware optimization pragmas
//See Impulse C documentation for more information
#pragma CO pipeline //pipeline the following code to increase throughput
#pragma CO set stageDelay 200 	//set max-stage-delay to 200 units.
								//Units roughly correspond to gate delays of target technology.
								//Max stage delay determines the maximum
								//frequency the logic can be clocked at.
					//Square complex number, add complex constant and check magnitude
					tmp = z_real;
					z_real = FXMUL(z_real, z_real);
					z_real = FXSUB(z_real, FXMUL(z_imag, z_imag));
					z_real = FXADD(z_real, c_real);
					z_imag = FXMUL(tmp, z_imag);
					//change *2 to <<1, which is more efficiently implemented in HW
					//z_imag = FXMUL(two,z_imag);
					z_imag = z_imag << 1; 
					z_imag = FXADD(z_imag, c_imag);
					tmp=FXMUL(z_real, z_real);
					result = FXADD(tmp, FXMUL(z_imag, z_imag));
					co_par_break(); //cycle boundary added to make logic fit within 10ns clock cycle. 
									//If removed timming will fail during implementation.
					k++;
				} while ((result < four) && (k < MAX_ITERATIONS)); //loop will exit if complex number diverges or max iterations is reached
				//Calculate pixel color
				//If number diverged then color assigned depends on number of iterations
                if (k != MAX_ITERATIONS) 
				{
					G = (k > 255) ? 255 : k;
					B = R = 0;
				}
                else //Else pixel is colored black
				{
					B = G = R = 0;	
				}
				
				//Assign color in Blue(8bit)-Green(8 bit)-Red(8 bit) format for storage in video memory.
                BGR = ((B << 16) & BLUEMASK) | ((G << 8) & GREENMASK) | (R & REDMASK); 
				
				//Write pixel value to stream, which will be read by the SW process on other end of stream.
				//co_stream_write() writes a data value to a stream.
				//Usage: co_stream_write(stream_name, variable_address, size_of_stream_element);
                co_stream_write(pixel_stream, &BGR, sizeof(co_uint24));
				c_real = FXADD(c_real, dx);
            }
            c_imag = FXSUB(c_imag, dy);
        }
		//co_stream_close() writes a end-of-stream(EOS) token to the selected stream
		//When the process on the other end of the stream reads the EOS, it can close its side of the stream
		//Usage: co_stream_close(stream_name);
		//Close pixel_stream after writing all pixel values to it.
        co_stream_close(pixel_stream);
    }
	//End-of-stream token has been read on config_stream, hence close HW side of config_stream
    co_stream_close(config_stream);
}

//Stream buffer lengths
#define CONFIG_BUFSIZE 10 	//Size is big enough for all image parameters to be buffered
#define STREAM_BUFSIZE 250	//Size is big enough for one image line to be buffered
							//Image size = 320 x 240

//Configuration function of CoDeveloper project
//Every CoDeveloper project must contain a single configuration function, which declares processes, 
//streams and connectivity in the system
void config_mand(void *arg)
{
    co_process mand_hw_proc, mand_sw_proc;	//system processes
    co_stream pixel_stream, config_stream;	//system streams

	//Declaration of streams
	//Usage : (co_stream)stream_name = co_stream_create("stream_name_for_sims", size_of_stream_element, no_of_stream_elements);
    pixel_stream = co_stream_create("pixel_stream", UINT_TYPE(24), STREAM_BUFSIZE);
    config_stream = co_stream_create("config_stream", INT_TYPE(32), CONFIG_BUFSIZE);

	//Declaration of processes
	//Usage : (co_process)process_names = co_process_create("process_name_for_sims", (co_function)function_name, number_of_arguments, arg1, arg2,....); 
    mand_hw_proc = co_process_create("mand_proc", (co_function)mand_accel_hw, 2, config_stream, pixel_stream);
    mand_sw_proc = co_process_create("test_proc", (co_function)mand_accel_sw, 2, config_stream, pixel_stream);

	//mand_hw_proc to be moved to hardware
	//Usage: co_process_config(process, attribute_to_change, attribute_value);
	//currently only attribute_to_change = co_loc, and attribute_value = PE0 is supported
    co_process_config(mand_hw_proc, co_loc, "PE0");
}

//Every CoDeveloper project must contain one co_initialize()
//which is called from main()
co_architecture co_initialize()     
{
	//Associate the above configuration function with a specific architecture.
	//The target architecture for compilation however is set in the project > options menu.
    return(co_architecture_create("mand", "generic", config_mand, NULL));
}
