/******************************************************************************
 *
 *     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
 *     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
 *     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
 *     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
 *     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
 *     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
 *     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
 *     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
 *     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
 *     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
 *     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
 *     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *     FOR A PARTICULAR PURPOSE.
 *     
 *     (c) Copyright 2005 Xilinx, Inc.
 *     All rights reserved.
 *
 ******************************************************************************/

//----------------------------------------------------------------------------
// Filename:     mand_accel_sw.c
// 
// Description: SW process that interacts with the HW Mandelbrot generation 
//				process to display the image generated by HW co-processor
//
// Design Notes:
//   
//-----------------------------------------------------------------------------
// Xilinx APD
// History:
// July 2005 - Created - Kunal Shenoy
//
//-----------------------------------------------------------------------------

#ifdef WIN32
	#include <windows.h>
	#include <sys/types.h>
#else  // ! WIN32
	#include <unistd.h>
#endif

#include <sys/stat.h>
#include <stdio.h>
#include "co.h"
#include "cosim_log.h"
#include "mand.h"

#ifdef IMPULSE_C_TARGET
	#include "bootload_basicgraphics.h"
	#include "xtft_main.h"
	#include "stop_watch.h"
#else
	//bitmap file for hardware-simulation image storage
	extern FILE * outfileHW;
#endif

//mand_accel_sw() is the process which feeds parameters to the HW mandelbrot 
//generation process. It also reads in the pixel values from the hardware process
//and plots them on the screen. 
//config_stream = stream over which mand_accel_sw passes parameters to 
//mand_accel_hw
//pixel_stream = stream over which mand_accel_sw reads in pixel values from 
//mand_accel_hw
void mand_accel_sw(co_stream config, co_stream pixel_stream)
{
    double xmax, xmin, ymax, ymin, dx, dy; //Image parameters
    co_int24 pixelValue; //color of pixel
    co_int32 data; //Image parameter in fixed-point format
    co_uint24 pixelCount = 0; //Keeps track of pixel completed in current line
#ifdef IMPULSE_C_TARGET
	int i = 0; //Pixel number in current line
	int j = 0; //Current line number
	uint32 graph_x = 61; //Starting x co-ordinate for speed graph
	uint32 time_line; //y-value of graph point
	co_error err;
#else
	co_uint24 byteValue; //stores pixel value for .bmp file
#endif
	static double theFactor = 0; //number by which the image parameters are varied
								 // to pan through image space

	//reset internal state of config stream
	//co_stream_open() resets the internal state of the stream
	//usage: co_stream_open(stream_name, O_RDONLY/0_WRONLY, size_of_stream_element);
    co_stream_open(config, O_WRONLY, INT_TYPE(32));

	//Image parameters
	//theFactor changes the image parameters to cycle through image space for the demo
    xmin = XL - theFactor;
    xmax = XH - theFactor;
    ymin = YL - theFactor;
    dx = (XH - XL) / XSIZE;
    dy = dx; //(YH-YL)/YSIZE;
    ymax = YL + (dy * YSIZE);
		
	if (theFactor < 0.036)
		theFactor += 0.003;
	else
		theFactor = 0;

	//Convert image parameters to fixed-point format and write to config_stream
	//SW process on other end of config_stream reads in image parameters and starts generating
	//Mandelbrot image
    data = (int32)(xmax * (1 << FRACBITS));
	//co_stream_write() writes a data value to a stream.
	//Usage: co_stream_write(stream_name, variable_address, size_of_stream_element);
    co_stream_write(config, &data, sizeof(co_int32));
    data = (int32)(xmin*(1 << FRACBITS));
    co_stream_write(config, &data, sizeof(co_int32));
    data = (int32)(ymax * (1 << FRACBITS));
    co_stream_write(config,&data,sizeof(co_int32));
    data = (int32)(ymin * (1 << FRACBITS));
    co_stream_write(config, &data, sizeof(co_int32));
    data = (int32)(dx * (1 << FRACBITS));
    co_stream_write(config, &data, sizeof(co_int32));
    data = (int32)(dy * (1 << FRACBITS));
    co_stream_write(config, &data, sizeof(co_int32));

	//co_stream_open() resets the internal state of the stream
	//usage: co_stream_open(stream_name, O_RDONLY/0_WRONLY, size_of_stream_element);
    co_stream_open(pixel_stream, O_RDONLY, UINT_TYPE(24));

#ifdef IMPULSE_C_TARGET
	//reset and start timer 1 of stopwatch to track completion time of line
	stop_watch(1,1); 
#endif

	//Loop continues till EOS(End Of Stream) token is received from mand_accel_hw process
	//EOS token by mand_accel_hw process indicates all pixels have been communicated
#ifdef IMPULSE_C_TARGET //If running on FPGA
	while (1)
	{
		// err = co_stream_read(pixel_stream, &pixelValue, sizeof(co_int32)); 
		// Use this macro instead of co_stream_read for faster APU throughput...
		HW_STREAM_READ(test_proc,pixel_stream,pixelValue,err);
		if (err != co_err_none)
			break;
		//plot pixel on screen
		XTft_SetPixel(&Tft, i+320, j+240, pixelValue);
		i++;
		//screen wrap around at end of horizontal line
		if (i >= XSIZE)
		{
			i = 0;
			j++;
			//calculate and plot speed graph 
			//Done for every line
			time_line = 435 - (40000 / stop_watch(4,1));
			XTft_SetPixel(&Tft, graph_x++, time_line, 0xFF00);
			stop_watch(1,1);
		}
#else 	//If running desktop simulation
	while (co_stream_read(pixel_stream, &pixelValue, sizeof(co_int32)) == co_err_none)
	{
		//Write output to HW result bitmap file 
        if (pixelCount % XSIZE == 0) 
		{
            printf("\rLine %d",pixelCount/XSIZE);
            fflush(stdout);
		}
		byteValue = 0;
        byteValue = (unsigned int) pixelValue & BLUEMASK;
        byteValue >>= 16;
        putc(byteValue, outfileHW);
        byteValue = (unsigned int) pixelValue & GREENMASK;
        byteValue >>= 8;
        putc(byteValue, outfileHW);
        byteValue = (unsigned int) pixelValue & REDMASK;
        putc(byteValue, outfileHW);
#endif
		pixelCount++;
    }
	//mand_accel_sw closes pixel_stream only after receiving a End Of Stream(EOS)
	//token from the HW process since the HW process is the writer to the stream.
	//co_stream_close() writes a end-of-stream(EOS) token to the selected stream
	//usage: co_stream_close(stream_name);
	co_stream_close(pixel_stream);

	//If runnning on the desktop simulation mode then mand_accel_sw closes config_stream
	//If running on FPGA, demo cycles through many images and hence stream is not closed.
	//The mand_accel_hw process on the other end sees the EOS token and closes its side of the stream.
	//co_stream_close() writes a end-of-stream(EOS) token to the selected stream
    //usage: co_stream_close(stream_name);
#ifndef IMPULSE_C_TARGET
	co_stream_close(config);
#endif
}

